// ============================================================================
//
// Copyright (c) 2010 - 2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================


package ch.niceideas.cometTennis;

import ch.niceideas.cometTennis.model.TennisGame;
import ch.niceideas.common.http.comet.CometDispatchClientException;
import ch.niceideas.common.http.comet.CometManager;
import ch.niceideas.common.http.comet.CometServerException;
import ch.niceideas.common.service.Service;
import ch.niceideas.common.utils.CommonBusinessException;
import org.apache.log4j.Logger;

import javax.annotation.Resource;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Main manager of the Tennis Game.
 */
@Service
public class TennisManager {

    private static final int EVENT_LOOP_DELAY = 40;

    private static final Logger logger = Logger.getLogger(TennisManager.class);

    @Resource
    private CometManager cometManager = null;
    
    private Map<String, TennisGame> games = new ConcurrentHashMap<String, TennisGame>();
    
    private Timer timer = new Timer();

    /**
     * Don't want this to be built outside
     * 
     * @throws CommonBusinessException
     */
    public TennisManager() throws CommonBusinessException {
        init();
    }

    private void init() throws CommonBusinessException {

        timer.schedule(new TimerTask() {
            @Override
            public void run() {
                try {
                    gamePeriod();
                } catch (Exception e) {
                    logger.debug (e, e);
                }
            }}, EVENT_LOOP_DELAY, EVENT_LOOP_DELAY);
    }
    
    private void gamePeriod() {
        Set<TennisGame> gameSet = getGameSet();
        for (TennisGame game : gameSet) {
            try {
                game.period();
            } catch (CometServerException e) {
                logger.error (e, e);
                
                CometDispatchClientException dispatchExp = null;
                if (e instanceof CometDispatchClientException) {
                    dispatchExp = ((CometDispatchClientException)e);
                } else if (e.getCause() != null && e.getCause() instanceof CometDispatchClientException) {
                    dispatchExp = ((CometDispatchClientException)e.getCause());
                }
                
                if (dispatchExp != null) {                    
                    String clientId = dispatchExp.getClientId();
                    
                    try {
                        if (clientId.equals(game.getFirstPlayerId())) {
                            cometManager.dispatch("GAME", game.getSecondPlayerId(), "NOTIFICATION:Game killed because other player left");
                        } else {
                            cometManager.dispatch("GAME", game.getFirstPlayerId(), "NOTIFICATION:Game killed because other player left");
                        }
                    } catch (CometServerException e1) {
                        logger.error(e1, e1);
                    }                    
                }
                
                games.remove(game.getFirstPlayerId());
                games.remove(game.getSecondPlayerId());
            }
        }
    }

    public Set<TennisGame> getGameSet() {
        return new HashSet<TennisGame>(games.values());
    }

    /**
     * Handle a message from UI.
     * 
     * @param clientId
     * @param message
     * @throws TennisException 
     */
    public void message(String clientId, String message) throws TennisException, CometDispatchClientException {
        
    	TennisGame game = games.get (clientId);
    	
        if (message.startsWith("INIT")) {
        	
        	if (game != null) {
        		throw new TennisException ("ClientId " + clientId + " is already in game " + game.getId());
        	}
        	
            game = new TennisGame(cometManager, this);
            game.addPlayer(clientId);
            games.put (clientId, game);
            try {
                cometManager.dispatch("GAME", clientId, "FIRST_PLAYER:" + game.getId());
            } catch (CometServerException e) {
                logger.error (e, e);
                throw new TennisException(e);
            }
                    
        } else if (message.startsWith("JOIN")) {
        	
        	if (game != null) {
        		throw new TennisException ("ClientId " + clientId + " is already in game " + game.getId());
        	}
            
            String gameIdStr = null;
            if (message.length() > 5) {
                gameIdStr = message.substring(5);
            } else {
                throw new TennisException("Expected valid gameId");
            }
            int gameId = Integer.parseInt(gameIdStr);
            
            Set<TennisGame> gameSet = getGameSet();
            boolean joinedGame = false;
            for (TennisGame otherGame : gameSet) {
                if (otherGame.getId() == gameId) {   
                    if (otherGame.hasFreeSlot()) {
                    	otherGame.addPlayer(clientId);
                        games.put (clientId, otherGame);
                        joinedGame = true;
                        try {
                            cometManager.dispatch("GAME", clientId, "SECOND_PLAYER");
                        } catch (CometServerException e) {
                            logger.error (e, e);
                            throw new TennisException(e);
                        }                        
                    }
                    break;
                }
            }
            if (!joinedGame) {
                message (clientId, "INIT");
            }
            
        } else {
                        
            if (game == null) {
                throw new TennisException ("No game found for clientId " + clientId);
            }
            
            if (message.startsWith("START_UP")) {        
                
                game.startUp (clientId);
                
            } else if (message.startsWith("START_DOWN")) {
                
                game.startDown (clientId);
            
            } else if (message.startsWith("STOP_UP")) {
                
                game.stopUp (clientId);
                
            } else if (message.startsWith("STOP_DOWN")) {
                
                game.stopDown (clientId);
                
            } else  {
                throw new TennisException("Unknown message : '" + message + "' from client " + clientId);
            }
        }
    }
    
  
}
