// ============================================================================
//
// Copyright (c) 2010 - 2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================


package ch.niceideas.cometTennis.model;

import java.util.Set;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import ch.niceideas.cometTennis.TennisException;
import ch.niceideas.cometTennis.TennisManager;
import ch.niceideas.common.http.comet.CometManager;
import ch.niceideas.common.http.comet.CometServerException;

/**
 * Represents a tennis game between and his 2 players.
 */
public class TennisGame {
    
    static final int BOARD_HEIGHT = 500;
    static final int BOARD_WIDTH = 800;
    
    static final int BALL_WIDTH = 10;
    
    static final int BALL_SPEED_X = 6;
    static final int INIT_BALL_SPEED_Y = 0;

    static final int PAD_HEIGHT = 100;
    static final int PAD_WIDTH_OFFSET = 50;
    
    static final int PAD_STEP = 5;        

    private final int id;
    private final CometManager cometManager;    
    private final TennisManager cometTennisManager;
    
    private TennisPlayer firstPlayer = null;
    private TennisPlayer secondPlayer = null;

    private TennisBall ball = null;
    
    public TennisGame(CometManager cometManager, TennisManager cometTennisManager) {
        this.cometManager = cometManager;
        this.cometTennisManager = cometTennisManager;
        
        this.id = generateId();
    }
    
    private int generateId() {
        boolean gameExist = true;
        int gameId = -1;
        while (gameExist) {
            gameExist = false;
            gameId = (int) (Math.random() * 100000);
            Set<TennisGame> games = this.cometTennisManager.getGameSet();
            for (TennisGame game : games) {
                if (game.getId() == gameId) {
                    gameExist = true;
                    break;
                }
            }
        }
        return gameId;
    }

    /**
     * 
     * @param playerId
     * @throws TennisException
     */
    public void addPlayer (String playerId) throws TennisException {
        if (firstPlayer == null) {
            firstPlayer = new TennisPlayer (playerId);
        } else if (secondPlayer == null) {
            secondPlayer = new TennisPlayer (playerId);
            ball = new TennisBall();
        } else {
            throw new TennisException ("Already two player in game");
        }
    }

    public void startUp(String clientId) throws TennisException {
        if (firstPlayer.matchesId (clientId)) {
            firstPlayer.startUp();
        } else if (secondPlayer.matchesId (clientId)) {
            secondPlayer.startUp();
        } else {
            throw new TennisException ("Unknown client "  + clientId);
        }
    }

    public void startDown(String clientId) throws TennisException {
        if (firstPlayer.matchesId (clientId)) {
            firstPlayer.startDown();
        } else if (secondPlayer.matchesId (clientId)) {
            secondPlayer.startDown();
        } else {
            throw new TennisException ("Unknown client "  + clientId);
        }
    }

    public void stopUp(String clientId) throws TennisException {
        if (firstPlayer.matchesId (clientId)) {
            firstPlayer.stopUp();
        } else if (secondPlayer.matchesId (clientId)) {
            secondPlayer.stopUp();
        } else {
            throw new TennisException ("Unknown client "  + clientId);
        }
    }

    public void stopDown(String clientId) throws TennisException {
        if (firstPlayer.matchesId (clientId)) {
            firstPlayer.stopDown();
        } else if (secondPlayer.matchesId (clientId)) {
            secondPlayer.stopDown();
        } else {
            throw new TennisException ("Unknown client "  + clientId);
        }
    }
    
    public void period() throws CometServerException {
        
        // move first player
        firstPlayer.period();
        
        // Move second player
        secondPlayer.period();      
        
        cometManager.dispatch("GAME", firstPlayer.getId(),  "POS:" + firstPlayer.getId()  + ":" + firstPlayer.getPos());
        cometManager.dispatch("GAME", secondPlayer.getId(), "POS:" + firstPlayer.getId()  + ":" + firstPlayer.getPos());
        cometManager.dispatch("GAME", firstPlayer.getId(),  "POS:" + secondPlayer.getId() + ":" + secondPlayer.getPos());
        cometManager.dispatch("GAME", secondPlayer.getId(), "POS:" + secondPlayer.getId() + ":" + secondPlayer.getPos());
        
        if (ball != null) {
            
            ball.period();            
            
            cometManager.dispatch("GAME", firstPlayer.getId(),  "BALL_POS_X:_:" + ball.getPosX());
            cometManager.dispatch("GAME", secondPlayer.getId(), "BALL_POS_X:_:" + ball.getPosX());
            cometManager.dispatch("GAME", firstPlayer.getId(),  "BALL_POS_Y:_:" + ball.getPosY());
            cometManager.dispatch("GAME", secondPlayer.getId(), "BALL_POS_Y:_:" + ball.getPosY());
            
            // Detect collisions with walls 
            GameMove winingPlayer = ball.detectWallCollisions();
            if (winingPlayer != null) {
                if (winingPlayer == GameMove.LEFT) {
                    cometManager.dispatch("GAME", firstPlayer.getId(),  "FIRST_PLAYER_WIN:_:_");
                    cometManager.dispatch("GAME", secondPlayer.getId(), "FIRST_PLAYER_WIN:_:_");
                } else {
                    cometManager.dispatch("GAME", firstPlayer.getId(),  "SECOND_PLAYER_WIN:_:_");
                    cometManager.dispatch("GAME", secondPlayer.getId(), "SECOND_PLAYER_WIN:_:_");
                }
            }
            
            // Detect collisions with players 
            
            // First player (Left)
            ball.detectCollisionLeftPlayer(firstPlayer);

            // Second player (Right)
            ball.detectCollisionRightPlayer(secondPlayer);
        }        
    }
    
    public String getFirstPlayerId() {
        return firstPlayer != null ? firstPlayer.getId() : null;
    }

    public String getSecondPlayerId() {
        return secondPlayer != null ? secondPlayer.getId() : null; 
    }

    public boolean hasFreeSlot() {
        return secondPlayer == null;
    }
    
    @Override
    public int hashCode() {
        return new HashCodeBuilder()
            .append(getFirstPlayerId())
            .append(getSecondPlayerId())
            .hashCode();
    }

    @Override
    public boolean equals(Object obj) {        
        if (this == obj) return true;
        if (obj == null) return false;
        if (getClass() != obj.getClass()) return false;
        TennisGame other = (TennisGame) obj;
        return new EqualsBuilder()
            .append(getFirstPlayerId(), other.getFirstPlayerId())
            .append(getSecondPlayerId(), other.getSecondPlayerId())
            .isEquals();
    }
    
    static enum PlayerMove {
        UP, DOWN, NONE;
    }
    
    static enum GameMove {
        LEFT, RIGHT;
    }

    public int getId() {
        return id;
    }     

}
