// ============================================================================
//
// Copyright (c) 2010 - 2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================



package ch.niceideas.common.http.comet;

import ch.niceideas.common.service.Service;
import ch.niceideas.common.utils.CommonBusinessException;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;

/**
 *   TODO Implement 
 *     DONE 1. automatic drop / recreate of iframe after 30 seconds
 *     2 storage of messages that might have been missed during the recreation
 *       server side and resent after reconnection.
 */
@Service
public class CometManager {

    private static final Logger logger = Logger.getLogger(CometManager.class);

    private static final int NUMBER_THREADS_FOR_SEND = 10;
    
    private Map<String, Map<String, CometHandle>> streamMap = new ConcurrentHashMap<String, Map<String, CometHandle>>();
    private ExecutorService pool = Executors.newFixedThreadPool(NUMBER_THREADS_FOR_SEND, new ThreadFactory() {        
        @Override
        public Thread newThread(Runnable r) {
            Thread thread = new Thread (r);
            thread.setDaemon(true);
            return thread;
        }
    });
    
    private Map<String, CometHandle> getHandleMapByStream (String streamId) {
        Map<String, CometHandle> handleMap = streamMap.get(streamId);
        if (handleMap == null) {
            synchronized (streamMap) {
                handleMap = streamMap.get(streamId);
                if (handleMap == null) {
                    handleMap = new ConcurrentHashMap<String, CometHandle>();
                    streamMap.put (streamId, handleMap);
                }
            }
        }
        return handleMap;
    }

    /**
     * Don't want this to be built outside
     * 
     * @throws CommonBusinessException
     */
    public CometManager() throws CommonBusinessException {

    }
    /**
     *
     * @param handle
     * @throws CometServerException
     */
    public void registerHandle (CometHandle handle) throws CometServerException {
        Map<String, CometHandle> handleMap = getHandleMapByStream (handle.getStreamId());
        if (handleMap.containsKey(handle.getClientId())) {
            throw new CometServerException ("Already has a client registered with ID " + handle.getClientId());
        }
        
        logger.info ("Registering handle : " + handle);
        handleMap.put (handle.getClientId(), handle);
    }
    
    /**
     * @param streamId
     * @param command
     */
    public void dispatchAll (String streamId, String command) {
        final DispatchCommand cmd = new DispatchCommand(streamId, command);
        pool.execute(new Runnable() {
            @Override
            public void run() {
                try {
                    cmd.dispatch();
                } catch (CometServerException e) {
                    logger.error (e, e);
                }
            }            
        });
    }
    
    /**
     * @param streamId
     * @param command
     * @throws CometServerException
     */
    private void doDispatchAll (String streamId, String command) throws CometServerException {
        Map<String, CometHandle> handleMap = getHandleMapByStream (streamId);
        if (StringUtils.isNotBlank(command)) {
            CometServerException exception = new CometServerException("");
            for (CometHandle handle : handleMap.values()) {
                try {
                    dispatchInternal(handleMap, command, handle);
                } catch (CometServerException e) {
                    logger.error (e, e);
                    exception.addUnderlyingException(e);
                }
            }
            exception.<CometServerException>throwIfAny();
        }
    }

    /**
     * @param streamId
     * @param clientId
     * @param command
     * @throws CometServerException
     */
    public void dispatch (String streamId, String clientId, String command) throws CometServerException {
        for (int i = 0; i < 100; i++) { // 10 tries, 100 ms sleep
            try {
                doDispatch(streamId, clientId, command);
                break;
            } catch (CometServerException e) {
                if (i >= 20) {
                    logger.error (e, e);
                    throw new CometDispatchClientException(streamId, clientId, e.getMessage(), e); 
                }
                logger.debug(e.getMessage());
                try {
                    Thread.sleep(100);
                } catch (InterruptedException e1) {
                    logger.error (e, e);
                }
            }
        }
    }
    
    /**
     * @param streamId
     * @param clientId
     * @param command
     * @throws CometServerException
     */
    private void doDispatch (String streamId, String clientId, String command) throws CometServerException {
        Map<String, CometHandle> handleMap = getHandleMapByStream (streamId);
        if (StringUtils.isNotBlank(command)) {
            CometHandle handle = handleMap.get(clientId);
            if (handle == null) {
                logger.debug("Found no handle with ID " + clientId);
                throw new CometServerException("Found no handle with ID " + clientId);
            }
            dispatchInternal(handleMap, command, handle);
        }
    }

    /**
     * 
     * @param command
     * @param handle
     * @throws CometServerException
     */
    private void dispatchInternal(Map<String, CometHandle> handleMap, String command, CometHandle handle)
            throws CometServerException {        
        try {
            
            if (command.trim().equals("KILL") || command.trim().equals("RECONNECT")) {
                handleMap.remove(handle.getClientId());
            }
            
            handle.sendMessage(command);
   
        } catch (CometStreamException e) {
            logger.error (e, e);
            handleMap.remove(handle.getClientId());
            handle.terminate();
            throw new CometServerException("Handle has been terminated");
        }
    }
    
    class DispatchCommand {
        
        private final String streamId;
        private final String clientId; 
        private final String command;
        
        public DispatchCommand(String streamId, String command) {
            super();
            this.streamId = streamId;
            this.clientId = null;
            this.command = command;
        } 
        
        public DispatchCommand(String streamId, String clientId, String command) {
            super();
            this.streamId = streamId;
            this.clientId = clientId;
            this.command = command;
        }       
        
        void dispatch () throws CometServerException {
            if (StringUtils.isBlank(clientId)) {
                doDispatchAll(streamId, command);
            } else {
                doDispatch(streamId, clientId, command);
            }
        }
    }
}
