// ============================================================================
//
// Copyright (c) 2010 - 2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================



package ch.niceideas.common.http.comet;

import ch.niceideas.common.http.comet.message.MessageHandler;
import ch.niceideas.common.http.comet.message.SystemMessageHandler;
import ch.niceideas.common.service.RegistryInitializationException;
import ch.niceideas.common.service.RegistryInitializer;
import ch.niceideas.common.service.StaticRegistry;
import ch.niceideas.common.utils.ResourceUtils;
import ch.niceideas.common.utils.StreamUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.concurrent.ConcurrentLinkedQueue;

public class CometMessageServlet extends HttpServlet {
        
    private static final Logger logger = Logger.getLogger(CometMessageServlet.class);
    
    private static final long serialVersionUID = -4717836700342947212L;

    private ConcurrentLinkedQueue<MessageHandler> handlersQueue = new ConcurrentLinkedQueue<MessageHandler>();

    private static StaticRegistry staticRegistry = null;

    static {
        if (staticRegistry == null) {
            synchronized (CometMessageServlet.class) {
                if (staticRegistry == null) {
                    try {
                        staticRegistry = (StaticRegistry) RegistryInitializer.init("ch.niceideas");
                    } catch (RegistryInitializationException e) {
                        logger.error (e, e);
                    }
                }
            }
        }
    }

    @Override
    public void init(ServletConfig config) throws ServletException {
        super.init (config);
        
        String messageHandlersConfig = config.getInitParameter("messageHandlers");
        if (StringUtils.isNotBlank(messageHandlersConfig)) {
            for (String messageHandlersClassString : messageHandlersConfig.split(",")) {
                try {
                    @SuppressWarnings("unchecked")
                    Class<MessageHandler> messageHandlersClass = (Class<MessageHandler>) Class.forName(messageHandlersClassString);
                    
                    MessageHandler handler = messageHandlersClass.newInstance();
                    handlersQueue.add(handler);
                    
                } catch (ClassNotFoundException e) {
                    logger.error (e, e);
                } catch (InstantiationException e) {
                    logger.error (e, e);
                } catch (IllegalAccessException e) {
                    logger.error (e, e);
                }
            }
        }
        
        // This one needs to be added last !
        handlersQueue.add(new SystemMessageHandler());
    }
    
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        doPost (req, resp);
    }
    
    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException {
        
        try {
            // Connection special case
            String message = req.getParameter("message");
            if (StringUtils.isNotBlank(message) && message.equalsIgnoreCase("connect")) {
                logger.info("");
                logger.info("Got incoming connection request with thread=" + Thread.currentThread().getName() + " / session=" + req.getSession().getId());
                InputStream is = ResourceUtils.getResourceAsStream("comet_init_connection.js");
                StreamUtils.copy(is, resp.getOutputStream());
                is.close();
                
                resp.flushBuffer();
                
            } else {
                
                
                String streamId = req.getParameter("streamId");
                if (StringUtils.isBlank(streamId)) { 
                    resp.sendError(500, "No 'streamId' provided.");
                    
                } else {
                    String clientId = req.getParameter("clientId");
                    
                    logger.info("");
                    logger.info("Got incoming message for streamId=" + streamId + " and from clientId=" + (StringUtils.isNotBlank(clientId) ? clientId : "(ALL)") + " with thread=" + Thread.currentThread().getName() + " / session=" + req.getSession().getId());
                    
                    if (StringUtils.isNotBlank(message)) { 
                        
                        logger.info(" - message : " + (message.length() > 10 ? message.substring(0,  10) : message));
                
                        for (MessageHandler handler : handlersQueue) {
                            if (handler.accept(req, streamId, clientId, message)) {
                                break;
                            }
                        }

                        // I don't know why but firefox doesn't like empty responses there ...
                        PrintWriter pw = new PrintWriter(resp.getOutputStream());
                        pw.write("<html><body></body></html>");
                        pw.close();
                        
                    } else {
                        resp.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "No 'message' provided.");
                    }
                }
            }

        
        } catch (CometBusinessException e) {
            logger.error (e, e);
            resp.setStatus(566);
            
            resp.setCharacterEncoding("UTF-8");
            resp.setContentType("text/plain");
            
            PrintWriter pw = new PrintWriter(resp.getOutputStream());
            pw.write(e.getMessage());
            pw.close();
            
        
        } catch (Exception e) {
            logger.error (e, e);
            resp.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e.getMessage());
        }
    }

}
