// ============================================================================
//
// Copyright (c) 2010 - 2011, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================


package ch.niceideas.common.service;

import ch.niceideas.common.utils.ReflectionUtils;
import javassist.*;
import org.apache.log4j.Logger;
import org.reflections.Reflections;

import javax.annotation.Resource;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * The RegistryInitializer is the most important component of SCIF - the Simple and Cute IoC Framework implemented
 * by this package.
 * <p />
 *
 * The RegistryInitializer is responsible for:
 * <p />
 *
 * <ul>
 * <li>Searching the classpath for classes declaring the @Service annotation</li>
 * <li>Injecting dependencies^in the various form supported by the IoC Framework :</li>
 *     <ul>
 *      <li><b>Field injection</b>. This is done using simply runtime reflection.</li>
 *      <li><b>Method (getter) injection</b>. This is done by generating dynamically a subclass that takes care of the
 *          <i>Lazy Loading</i> of the dependency.
 *      </li>
 *      </ul>
 * </ul>
 */
public class RegistryInitializer {

    private static final Logger logger = Logger.getLogger(RegistryInitializer.class);

    /**
     * Initialize a registry from the root package (prefix) given as argument.
     * <br />
     * The RegistryInitializer will search for classes annotated with @Service and add them to the returned registry.
     *
     * @param rootPackage
     * @return A Registry containing all discovered services
     * @throws RegistryInitializationException in case of any error
     */
    public static Registry init(String rootPackage) throws RegistryInitializationException {

        try {
            Reflections reflections = new Reflections(rootPackage);

            Set<Class<?>> annotated = reflections.getTypesAnnotatedWith(Service.class);

            // Create service wrappers
            Map<Class<?>, ServiceWrapper> wrappers = initServiceWrappers(annotated);

            // Build set of methods and fields to be analyzed
            analyzeWrappers(wrappers);

            // Now the complicated part, overwrite getters / methods using Javassist,
            // dynamically creating a subclass with bytecode manipulation
            enhancePropertyGetters(wrappers);

            StaticRegistry registry = new StaticRegistry();

            // Instantiate all the services and store them in Registry with class name as service name
            initializeRegistry(wrappers, registry);

            // Then do the easy thing : inject service on dependencies expressed on fields
            proceedFieldInjection(wrappers, registry);

            return registry;

        } catch (ReflectiveOperationException | NotFoundException | CannotCompileException e) {
            logger.error (e, e);
            throw new RegistryInitializationException (e.getMessage(), e);
        }
    }

    private static void proceedFieldInjection(Map<Class<?>, ServiceWrapper> wrappers, StaticRegistry registry) throws NoSuchFieldException, IllegalAccessException {
        for (ServiceWrapper wrapper : wrappers.values()) {

            String serviceName = wrapper.getServiceName();
            Object service = registry.getService(serviceName);

            for (Field field : wrapper.getFieldsToInit()) {

                Class<?> fieldType = field.getType();

                String fieldServiceName = fieldType.getCanonicalName();
                ReflectionUtils.setFailsafeFieldValue(field, service, registry.getService(fieldServiceName));

                logger.info ("Injected field '" + field.getName() + "' on service '" + serviceName + "'");
            }
        }
    }

    private static void initializeRegistry(Map<Class<?>, ServiceWrapper> wrappers, StaticRegistry registry) throws InstantiationException, IllegalAccessException, InvocationTargetException {
        for (ServiceWrapper wrapper : wrappers.values()) {

            String serviceName = wrapper.getServiceName();

            Class<?> clazz = wrapper.getServiceClass();

            Object service = clazz.newInstance();

            registry.storeService(serviceName, Registry.Scope.APPLICATION, service);

            // Try to inject registry
            Method setter = null;
            try {
                setter = ReflectionUtils.getSetter(clazz, "registry", Registry.class);
                setter.invoke(clazz, new Object[] { registry });
            } catch (NoSuchMethodException e) {
                logger.debug(e, e);
            }

            logger.info ("Storing service : '" + serviceName + "'");
        }
    }

    private static void enhancePropertyGetters(Map<Class<?>, ServiceWrapper> wrappers) throws NotFoundException, CannotCompileException, ClassNotFoundException {
        ClassPool.doPruning = true;

        ClassPool pool = ClassPool.getDefault();
        pool.appendClassPath(new LoaderClassPath(Thread.currentThread().getContextClassLoader()));

        for (ServiceWrapper wrapper : wrappers.values()) {

            if (wrapper.getMethodsToEnhance().size() > 0) {

                CtClass superClazz = pool.get(wrapper.getServiceName());

                // Unfortunately I need to go with the sub-class approach
                // I cannot change the original class since it has already been loaded and
                // javassist cannot change a class that is already loaded (that would require
                // changing linking and javassist cannot do that)

                CtClass clazz = pool.makeClass(wrapper.getServiceName() + "$javassist_sub");
                clazz.stopPruning(true);

                clazz.setSuperclass(superClazz);
                clazz.setModifiers(Modifier.PUBLIC);

                // Just a marker to avoid modifying twice the same class
                CtField modificationMarkerField = null;
                try {
                    modificationMarkerField = clazz.getDeclaredField("modificationMarker");
                } catch (NotFoundException e) {
                    // ignored
                }
                if (modificationMarkerField == null) {

                    logger.info("Processing modifications on class " + clazz.getName());

                    CtClass booleanClass = pool.get("java.lang.Boolean");
                    modificationMarkerField = new CtField(booleanClass, "modificationMarker", clazz);
                    modificationMarkerField.setModifiers(Modifier.setPrivate(Modifier.STATIC));
                    clazz.addField(modificationMarkerField);

                    // Add registry on class if not already one. The field might already have been
                    // added on a parent class. If this is the case, don't add it again
                    injectRegistryField(pool, clazz);

                    // Proceed with method modification
                    for (Method method : wrapper.getMethodsToEnhance()) {

                        // Various cases :

                        // 1. Method doesn't have the form of a getter
                        if (!method.getName().startsWith("get")) {

                            // => Just override method so it returns the service from registry
                            overrideMethod(clazz, method);
                        }

                        // 2. Method is a getter
                        else {

                            try {
                                Method setter = ReflectionUtils.getSetter(wrapper.getServiceClass(),
                                        ReflectionUtils.getPropertyName(method),
                                        method.getReturnType());

                                // 2.2 setter is found
                                // => Lazy loading : use underlying field as cache.
                                // If it it set, do nothing.
                                // If is is null, look for service and attach it
                                // Needs to override the getter for doing my business and
                                // delegate to the setter for setting the field
                                overrideGetter(clazz, method, setter);

                            } catch (NoSuchMethodException e) {

                                logger.debug (e, e);

                                // 2.1 No setter could be found
                                // => Just rewrite method so it returns the service from the registry
                                overrideMethod(clazz, method);
                            }
                        }
                    }
                }

                // make new subclass available to classloader
                pool.toClass(clazz, Thread.currentThread().getContextClassLoader(), null);
                clazz.stopPruning(false);

                // use the new subclass instead of the original class from now on
                Class<?> subClazz = Class.forName(clazz.getName());
                wrapper.overrideClass(subClazz);
            }
        }
    }

    private static void analyzeWrappers(Map<Class<?>, ServiceWrapper> wrappers) throws RegistryInitializationException {
        for (Class<?> clazz : wrappers.keySet()) {

            ServiceWrapper wrapper = wrappers.get(clazz);

            Collection<Method> meths = ReflectionUtils.getMethods(clazz);
            for (Method method : meths) {

                Resource resAnnot = method.getAnnotation(Resource.class);
                if (resAnnot != null) {

                    Class<?> retType = method.getReturnType();

                    ServiceWrapper retWrap = wrappers.get(retType);
                    if (retWrap == null) {
                        throw new RegistryInitializationException ("Service '" + clazz + "' depends on unknown service '" + retType + "'");
                    }

                    wrapper.addMethodtoEnhance(method);
                }
            }

            Collection<Field> flds = ReflectionUtils.getFields(clazz);
            for (Field field : flds) {
                Resource resAnnot = field.getAnnotation(Resource.class);
                if (resAnnot != null) {

                    Class<?> retType = field.getType();

                    ServiceWrapper retWrap = wrappers.get(retType);
                    if (retWrap == null) {
                        throw new RegistryInitializationException ("Service '" + clazz + "' depends on unknown service '" + retType + "'");
                    }

                    wrapper.addFieldToInit(field);
                }
            }
        }
    }

    private static Map<Class<?>, ServiceWrapper> initServiceWrappers(Set<Class<?>> annotated) {
        Map<Class<?>, ServiceWrapper> wrappers = new HashMap<Class<?>, ServiceWrapper>();
        for (Class<?> clazz : annotated) {
            logger.info ("Found service class '" + clazz + "'");
            wrappers.put (clazz, new ServiceWrapper(clazz));
        }
        return wrappers;
    }

    private static void overrideGetter(CtClass clazz, Method getter, Method setter) throws CannotCompileException {
        String targetService = getter.getReturnType().getCanonicalName();
        CtMethod newMethod = CtNewMethod.make(
                "public " + targetService + " " + getter.getName() + "() { \n" +
                        "" +
                        "    " + targetService + " retObject =  super." + getter.getName() + "(); " +
                        "    if (retObject == null) {" +
                        "         retObject =  (" + targetService + ") getRegistry().getService(\"" + targetService + "\"); " +
                        "         super." + setter.getName() + "(retObject);" +
                        "    }" +
                        "    return retObject;" +
                        "" +
                        "}",
                clazz);
        clazz.addMethod(newMethod);
    }

    private static void overrideMethod(CtClass clazz, Method method) throws CannotCompileException {
        String targetService = method.getReturnType().getCanonicalName();
        CtMethod newMethod = CtNewMethod.make(
                    "public " + targetService + " " + method.getName() + "() { \n" +
                         "" +
                         "    " + targetService + " retObject =  (" + targetService + ") getRegistry().getService(\"" + targetService + "\"); " +
                         "    return retObject;" +
                         "" +
                         "}",
                clazz);
        clazz.addMethod(newMethod);
    }

    /**
     * Inject a field to store the registry in the target clazz as well as a getter to retrieve that registry.
     *
     * @param pool the Javassist ClassPool to be used
     * @param clazz the class to be modified this way
     * @throws NotFoundException
     * @throws CannotCompileException
     */
    public static void injectRegistryField(ClassPool pool, CtClass clazz) throws NotFoundException, CannotCompileException {
        CtField registryField = null;
        try {
            registryField = clazz.getField("registry");
        } catch (NotFoundException e) {
            // ignored
        }
        if (registryField == null) {
            CtClass registryClass = pool.get(Registry.class.getName());
            registryField = new CtField(registryClass, "registry", clazz);
            registryField.setModifiers(Modifier.setPrivate(Modifier.STATIC));
            clazz.addField(registryField, "null");

            CtMethod registryGetter = CtNewMethod.getter("getRegistry", registryField);
            registryGetter.setModifiers(Modifier.PUBLIC);
            clazz.addMethod(registryGetter);

            CtMethod registrySetter = CtNewMethod.make(
                    "public static void setRegistry (" + Registry.class.getName() + " holder) { " +
                         "   registry = holder; " +
                         "} ",
                    clazz);
            clazz.addMethod(registrySetter);
        }
    }
}
