// ============================================================================
//
// Copyright (c) 2010 - 2011, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================

package ch.niceideas.common.service;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpSession;
import java.util.HashMap;
import java.util.Map;

/**
 * This is a specific Registry to be used within web applications.
 * <p />
 * 
 * It supports the various scopes the following ways :
 * <p />
 * 
 * <ul>
 * <li><b>APPLICATION</b> : the service is stored in a static map</li>
 * <li><b>CONTEXT</b> : the service is stored in the web application context</li>
 * <li><b>USER</b> : the service is stored in the HTTP session</li>
 * <li><b>LOCAL</b> : the service is stored in a ThreadLocal</li>
 * </ul>
 */
public class WebRegistry extends StaticRegistry implements Registry {

    //private static final Logger logger = Logger.getLogger(WebRegistry.class);

    private final HttpSession session;
    private final ServletContext context;

    private ThreadLocal<Map<String, Object>> localServices = new ThreadLocal<Map<String, Object>>();

    /**
     * Create an empty WebRegistry from an HTTPSession
     * 
     * @param session the httpSession to store service when using scope USER
     */
    public WebRegistry(HttpSession session) {
        this.session = session;
        if (session != null) {
            context = session.getServletContext();
        } else {
            context = null;
        }        
    }

    /**
     * Create an empty WebRegistry from an HTTPSession
     *
     * @param context the servletContext to use when storing services in scope CONTEXT
     */
    public WebRegistry(ServletContext context) {
        session = null;
        this.context = context;
    }

    /**
     * Create an empty WebRegistry from an HTTPSession
     *
     * @param session the httpSession to store service when using scope USER
     * @param context the servletContext to use when storing services in scope CONTEXT
     */
    public WebRegistry(HttpSession session, ServletContext context) {
        this.session = session;
        if (context != null) {
            this.context = context;
        } else {
            this.context = session.getServletContext();
        }
    }

    /**
     * {@inheritDoc}
     */
    public Object getService(Object key) {

        Object retService = null;

        // 1. First try to get the service from local scope
        Map<String, Object> localServicesMap = localServices.get();
        if (retService == null && localServices != null && localServicesMap != null) {
            retService = localServicesMap.get(key.toString());
        }

        // 2. Try to get service from user scope
        if (retService == null && session != null) {
            retService = session.getAttribute(key.toString());
        }

        // 3. Try to get the service from the servlet scope
        if (retService == null && context != null) {
            retService = context.getAttribute(key.toString());
        }

        // 4. If none is found, get it from the application scope
        if (retService == null) {
            retService = super.getService(key);
        }

        return retService;
    }

    /**
     * {@inheritDoc}
     */
    public void storeService(Object key, Scope scope, Object service) {
        if (scope.equals(Scope.APPLICATION)) {
            super.storeService(key, Scope.APPLICATION, service);

        } else if (scope.equals(Scope.CONTEXT)) {
            if (context == null) {
                throw new IllegalStateException(
                        "Have been asked to store service in the 'CONTEXT' scope but the context is null.");
            }
            context.setAttribute(key.toString(), service);

        } else if (scope.equals(Scope.LOCAL)) {
            Map<String, Object> localServicesMap = localServices.get();
            if (localServicesMap == null) {
                localServicesMap = new HashMap<String, Object>();
                localServices.set(localServicesMap);
            }
            localServicesMap.put(key.toString(), service);

        } else {
            if (session == null) {
                throw new IllegalStateException("No specific session available in call context");
            }
            session.setAttribute(key.toString(), service);
        }
    }

}
