// ============================================================================
//
// Copyright (c) 2010 - 2011, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================

package ch.niceideas.common.utils;

import org.apache.log4j.Logger;
import sun.reflect.FieldAccessor;
import sun.reflect.ReflectionFactory;

import java.lang.reflect.*;
import java.util.*;

/**
 * The ReflectionUtils class offers various utilities on top to the java reflection API.
 */
public class ReflectionUtils {

    private static final Logger logger = Logger.getLogger(ReflectionUtils.class);

    private static final ReflectionFactory reflectionFactory = ReflectionFactory.getReflectionFactory();

    private static final String MODIFIERS_FIELD = "modifiers";

    /**
     * Invoke the specified {@link Method} against the supplied target object with no arguments. The target object
     * can be <code>null</code> when invoking a static {@link Method}.
     * 
     * @param method the method to invoke
     * @param target the target object to invoke the method on
     * @return the invocation result, if any
     * @see #invokeMethod(Method, Object, Object[])
     */
    public static Object invokeMethod(Method method, Object target) {
        return invokeMethod(method, target, null);
    }

    /**
     * Invoke the specified {@link Method} against the supplied target object with the supplied arguments. The
     * target object can be <code>null</code> when invoking a static {@link Method}.
     * 
     * @param method the method to invoke
     * @param target the target object to invoke the method on
     * @param args the invocation arguments (may be <code>null</code>)
     * @return the invocation result, if any
     */
    public static Object invokeMethod(Method method, Object target, Object[] args) {
        try {
            return method.invoke(target, args);
        } catch (Exception ex) {
            logger.error(ex, ex);
            throw new CommonRTException(ex);
        }
    }

    /**
     * This is used to get a setter list from the class given as argument. Properties mapped to transient fields can
     * be skipped or not depending on the value of the <i>useTransient</i> argument.<br />
     * Only the methods from the class itself are taken in consideration, not those from the parent class.
     * 
     * @param clazz the class to search setters in
     * @param useTransients should transients fields be considered or not
     * @return the list of Setters found on clazz
     */
    public static List<Method> getSetters(Class<?> clazz, boolean useTransients) {

        List<Method> retMethodList = new ArrayList<Method>();

        Method[] methods = clazz.getDeclaredMethods();
        AccessibleObject.setAccessible(methods, true);
        for (Method method : methods) {
            Method m = method;
            if (
            // 1. We consider property getters only
            m.getName().startsWith("set")

            // 2. Avoid CGLIB hacks
                    && (m.getName().indexOf('$') == -1)

                    // 3. Should we use transients
                    && (useTransients || !isTransient(m))

                    // 4. No static
                    && (!Modifier.isStatic(m.getModifiers()))

                    // 5. Getters have one single parameter
                    && (m.getParameterTypes() != null && m.getParameterTypes().length == 1)) {

                retMethodList.add(m);
            }
        }

        return retMethodList;
    }

    /**
     * This is used to get a getter list from the class given as argument. Properties mapped to transient fields can
     * be skipped or not depending on the value of the <i>useTransient</i> argument.<br />
     * Only the methods from the class itself are taken in consideration, not those from the parent class.
     *
     * @param clazz the class to search getters in
     * @param useTransients should transients fields be considered or not
     * @return the list of Getters found on clazz
     */
    public static List<Method> getGetters(Class<?> clazz, boolean useTransients) {

        List<Method> retMethodList = new ArrayList<Method>();

        Method[] methods = clazz.getDeclaredMethods();
        AccessibleObject.setAccessible(methods, true);
        for (Method method : methods) {
            Method m = method;
            if (
            // 1. We consider property getters only
            (m.getName().startsWith("get") || m.getName().startsWith("is"))

            // 2. Avoid CGLIB hacks
                    && (m.getName().indexOf('$') == -1)

                    // 3. Should we use transients
                    && (useTransients || !isTransient(m))

                    // 4. No static
                    && (!Modifier.isStatic(m.getModifiers()))

                    // 5. Getters have no parameter
                    && (m.getParameterTypes() == null || m.getParameterTypes().length == 0)) {

                retMethodList.add(m);
            }
        }

        Collections.sort(retMethodList, new Comparator<Method>() {
            @Override
            public int compare(Method o1, Method o2) {
                return o1.getName().compareTo(o2.getName());
            }            
        });
        return retMethodList;
    }

    private static boolean isTransient(Method m) {

        // 1. get field name
        String fieldName = getPropertyName(m);

        if (fieldName == null) {
            return false;
        }

        // 2. get field
        Field correspondingField = null;
        try {
            correspondingField = m.getDeclaringClass().getDeclaredField(fieldName);
        } catch (SecurityException e) {
            return false;
        } catch (NoSuchFieldException e) {
            correspondingField = findField(fieldName, m.getDeclaringClass());

        }

        if (correspondingField == null) {
            return false;
        }

        return Modifier.isTransient(correspondingField.getModifiers());
    }

    /**
     * Return the name of a property identified either by a getter or a setter
     *
     * @param m the getter or setter
     * @return the name of the corresponding Java Propety
     */
    public static String getPropertyName(Method m) {
        String fieldName = null;
        if (m.getName().startsWith("get")) {

            fieldName = m.getName().substring(3, 4).toLowerCase() + m.getName().substring(4);

        } else if (m.getName().startsWith("is")) {

            fieldName = m.getName().substring(2, 3).toLowerCase() + m.getName().substring(3);

        } else if (m.getName().startsWith("set")) {

            fieldName = m.getName().substring(3, 4).toLowerCase() + m.getName().substring(4);
        }
        return fieldName;
    }

    private static Field findField(String fieldName, Class<?> clazz) {
        for (Field field : clazz.getDeclaredFields()) {
            if (field.getName().equals(fieldName)) {
                return field;
            }
        }
        return null;
    }

    /**
     * This is used to get the type of the value following a java bean like path on the object given as argument.
     * The path allows to browse an object composition graph.
     * <p />
     * 
     * For instance :
     * <p />
     * 
     * <ul>
     * <li><b>To get the object itself : </b> '<code>.</code>'</li>
     * <li><b>To get the property mapProp on the object</b> <code>this.getMaProp()</code> <b>:</b> '
     * <code>maProp</code>'</li>
     * <li><b>To get a property in the object returned by getMaProp() : </b>
     * <code>this.getMaProp().getOtherProp()</code> <b>:</b> '<code>maProp.otherProp</code>'</li>
     * </ul>
     * 
     * @param source the source class to start from when evaluation the path
     * @param path the property path to be used
     * @return the type of the result property
     */
    public static Class<?> getTypeOfPath(Class<?> source, String path) {

        StringTokenizer tk = new StringTokenizer(path, ".", false);
        Class<?> type = source;

        while (tk.hasMoreTokens() && type != null) {
            String nextPath = tk.nextToken();

            try {
                Method getter = ReflectionUtils.getGetter(type, nextPath);

                if (getter == null) {
                    throw new CommonRTException("Could not find any getter for property '" + nextPath + "' on class '"
                            + type + "'");
                }

                type = getter.getReturnType();

            } catch (IllegalArgumentException | SecurityException | NoSuchMethodException e) {
                logger.error(e, e);
                throw new CommonRTException(e);
            }
        }

        return type;
    }

    /**
     * This is used to get the value following a java bean like path on the object given as argument. The path
     * allows to browse an object composition graph.
     * <p />
     * 
     * For instance :
     * <p />
     * 
     * <ul>
     * <li><b>To get the object itself : </b> '<code>.</code>'</li>
     * <li><b>To get the property mapProp on the object</b> <code>this.getMaProp()</code> <b>:</b> '
     * <code>maProp</code>'</li>
     * <li><b>To get a property in the object returned by getMaProp() : </b>
     * <code>this.getMaProp().getOtherProp()</code> <b>:</b> '<code>maProp.otherProp</code>'</li>
     * </ul>
     *
     * @param param the source class to start from when evaluation the path
     * @param path the property path to be used
     * @return the value of the resulting property
     */
    public static Object getValueFromPath(Object param, String path) {

        StringTokenizer tk = new StringTokenizer(path, ".", false);
        Object current = param;

        while (tk.hasMoreTokens() && current != null) {
            String nextPath = tk.nextToken();

            try {
                Method getter = ReflectionUtils.getGetter(current.getClass(), nextPath);

                if (getter == null) {
                    throw new CommonRTException("Could not find any getter for property '" + nextPath + "' on class '"
                            + current.getClass() + "'");
                }

                current = getter.invoke(current, new Object[0]);

            } catch (IllegalArgumentException | IllegalAccessException | InvocationTargetException |
                     SecurityException | NoSuchMethodException  e) {
                logger.error(e, e);
                throw new CommonRTException(e);
            }
        }

        return current;
    }

    /**
     * Get the set of fields declared by a class or all its super-classes
     *
     * @param clazz the clazz to search in
     * @return the collection of discovered declared fields
     * @throws SecurityException in case anything goes wrong in regards to security or policy
     */
    public static Collection<Field> getFields(Class<?> clazz) {

        Collection<Field> retFields = new ArrayList<Field>();

        // 1. First try to find field the most accurate possible way
        Field searchedField = null;
        Class<?> current = clazz;
        while (current != null && !current.equals(Object.class)) {

            for (Field field : current.getDeclaredFields()) {
                if (!retFields.contains(field)) {
                    retFields.add (field);
                }
            }

            current = current.getSuperclass();
        }

        return retFields;
    }

    /**
     * Used to get the field 'fieldName' on the class given as argument. <b>This method allows to get the asked
     * field in the class itself or any of its parent class.</b>
     * 
     * @param clazz the class to search the field in
     * @param fieldName the name of the field to search
     * @return the matching field if found
     * @throws SecurityException
     * @throws NoSuchFieldException
     */
    public static Field getField(Class<?> clazz, String fieldName) throws NoSuchFieldException {

        // 1. First try to find field the most accurate possible way
        Field searchedField = null;
        Class<?> current = clazz;
        while (current != null && !current.equals(Object.class)) {
            try {
                searchedField = current.getDeclaredField(fieldName);
            } catch (NoSuchFieldException e) {
                logger.debug(e, e);
            }
            if (searchedField != null) {
                break;
            }
            current = current.getSuperclass();
        }

        if (searchedField != null) {
            AccessibleObject.setAccessible(new Field[] { searchedField }, true);
        } else {
            throw new NoSuchFieldException("No field '" + fieldName + "' found on class '" + clazz + "'");
        }

        return searchedField;
    }

    /**
     * This is used to know whether or not that target class get has a method named 'methodName'
     * <p />
     * The searched method is looked for in the class itself and in each of its parent class should it not be found
     * there.
     * 
     * @param clazz the class to search the method in
     * @param methodName the name of the method to search for
     * @param arguments the arguments of the method, required to identify it precisely
     * @return true if the method exists, false otherwise
     * @throws SecurityException
     */
    public static boolean hasMethod(Class<?> clazz, String methodName, Class<?>[] arguments) throws SecurityException {

        // 1. First simply try with getMethod

        // 2. If method could not be found using the above way, try with the below
        Method searchedMethod = getMethodInternal(clazz, methodName, arguments);

        return (searchedMethod != null);
    }

    /**
     * This is used to get the method named 'methodName' and accepting the given parameters on the class given as
     * argument.
     * <p />
     * The searched method is looked for in the class itself and in each of its parent class should it not be found
     * there.
     *
     * @param clazz the class to search the method in
     * @param methodName the name of the method to search for
     * @param arguments the arguments of the method, required to identify it precisely
     * @return the method if found, otherwise an exception is thrown
     * @throws SecurityException
     * @throws NoSuchMethodException
     */
    public static Method getMethod(Class<?> clazz, String methodName, Class<?>[] arguments) throws SecurityException,
            NoSuchMethodException {

        // 1. First simply try with getMethod
        Method searchedMethod = getMethodInternal(clazz, methodName, arguments);

        if (searchedMethod != null) {
            AccessibleObject.setAccessible(new Method[] { searchedMethod }, true);
        } else {
            throw new NoSuchMethodException("No method '" + methodName + "' found on class '" + clazz + "'");
        }

        return searchedMethod;
    }

    private static Method getMethodInternal(Class<?> clazz, String methodName, Class<?>[] arguments) {
        // 2. If method could not be found using the above way, try with the below
        Method searchedMethod = null;
        Class<?> current = clazz;
        while (current != null && !current.equals(Object.class)) {
            try {
                searchedMethod = current.getDeclaredMethod(methodName, arguments);
            } catch (NoSuchMethodException e) {
                logger.debug(e, e);
            }
            if (searchedMethod != null) {
                break;
            }
            current = current.getSuperclass();
        }

        // 2. If none could be found the way above, try to find the closest one
        if (searchedMethod == null) {
            current = clazz;
            while (current != null && !current.equals(Object.class)) {
                Method[] methods = current.getDeclaredMethods();
                for (Method method : methods) {
                    if (method.getName().equals(methodName) && method.getParameterTypes().length == arguments.length) {
                        searchedMethod = method;
                    }
                }
                if (searchedMethod != null) {
                    break;
                }
                current = current.getSuperclass();
            }
        }
        return searchedMethod;
    }

    /**
     * Get the collection of methods declared by the class given as argument
     *
     * @param clazz the class to list the methods of
     * @return the collection of methods declared by the class
     * @throws SecurityException
     */
    public static Collection<Method> getMethods(Class<?> clazz) throws SecurityException{

        Collection<Method> retMethods = new ArrayList<Method>();

        Class<?> current = clazz;

        current = clazz;
        while (current != null && !current.equals(Object.class)) {
            Method[] methods = current.getDeclaredMethods();
            for (Method method : methods) {
                if (!retMethods.contains(method)) {
                    retMethods.add(method);
                }
            }
            current = current.getSuperclass();
        }

        return retMethods;
    }

    /**
     * This is used to get the setter for the property 'propertyName' on the class given as argument.
     * <p />
     * The searched method is looked for in the class itself and in each of its parent class should it not be found
     * there.
     * 
     * @param clazz the class to search the setter in
     * @param propertyName the name of the property for which the setter should be returned
     * @param parameterClass the type of the parameter accepted by the setter
     * @return the setter if found, otherwise a NoSuchMethodException is thrown
     * @throws SecurityException
     * @throws NoSuchMethodException
     */
    public static Method getSetter(Class<?> clazz, String propertyName, Class<?> parameterClass)
            throws SecurityException, NoSuchMethodException {
        String methodName = "set" + firstLetterUpperCase(propertyName);
        return getMethod(clazz, methodName, new Class<?>[] { parameterClass });
    }

    /**
     * This is used to get the getter for the property 'propertyName' on the class given as argument.
     * <p />
     * The searched method is looked for in the class itself and in each of its parent class should it not be found
     * there.
     *
     * @param clazz the class to search the getter in
     * @param propertyName the name of the property for which the getter should be returned
     * @return the setter if found, otherwise a NoSuchMethodException is thrown
     * @throws SecurityException
     * @throws NoSuchMethodException
     */
    public static Method getGetter(Class<?> clazz, String propertyName) throws SecurityException, NoSuchMethodException {
        String methodName = "get" + firstLetterUpperCase(propertyName);
        try {
            return getMethod(clazz, methodName, new Class<?>[0]);
        } catch (NoSuchMethodException e) {
            logger.debug(e);
            methodName = "is" + firstLetterUpperCase(propertyName);
            return getMethod(clazz, methodName, new Class<?>[0]);
        }
    }

    /**
     * This is used to know whether or not the target class has a property with name given as argument
     * <p />
     * The searched property is looked for in the class itself and in each of its parent class should it not be
     * found there.
     *
     * @param clazz the class to search the property in
     * @param propertyName the name of the property to be searched
     * @return true if the property exists, false otherwise
     */
    public static boolean hasProperty(Class<?> clazz, String propertyName) {
        String methodName = "get" + firstLetterUpperCase(propertyName);
        boolean hasGetMethod = hasMethod(clazz, methodName, new Class<?>[0]);
        if (hasGetMethod) {
            return true;
        }

        methodName = "is" + firstLetterUpperCase(propertyName);
        return hasMethod(clazz, methodName, new Class<?>[0]);
    }


    /**
     * Put the first letter to lower case in the string given as argument and return the new string.
     *
     * @param original
     * @return Same as propertyName but with first letter lower case.
     */
    public static String firstLetterLowerCase(String original) {
        char[] charArray = original.toCharArray();
        charArray[0] = String.valueOf(charArray[0]).toLowerCase().charAt(0);
        return String.valueOf(charArray);
    }

    /**
     * Put the first letter to upper case in the string given as argument and return the new string.
     *
     * @param original
     * @return Same as propertyName but with first letter upper case.
     */
    public static String firstLetterUpperCase(String original) {
        char[] charArray = original.toCharArray();
        charArray[0] = String.valueOf(charArray[0]).toUpperCase().charAt(0);
        return String.valueOf(charArray);
    }

    /**
     * A utility method one can use to set a field with any kind of modifier making it illegal to actually set a
     * value (i.e private, final, static final, etc.)
     * 
     * @param field the field to be set
     * @param object the object on which to se tthe field
     * @param value the value to be set to the field
     * @throws NoSuchFieldException if the field doesn't exist
     * @throws IllegalAccessException
     */
    public static void setFailsafeFieldValue(Field field, Object object, Object value) throws NoSuchFieldException,
            IllegalAccessException {

        // we mark the field to be public
        field.setAccessible(true);

        // next we change the modifier in the Field instance to
        // not be final anymore, thus tricking reflection into
        // letting us modify the static final field
        Field modifiersField = Field.class.getDeclaredField(MODIFIERS_FIELD);
        modifiersField.setAccessible(true);
        int modifiers = modifiersField.getInt(field);

        // blank out the final bit in the modifiers int
        modifiers &= ~Modifier.FINAL;
        modifiersField.setInt(field, modifiers);

        FieldAccessor fa = reflectionFactory.newFieldAccessor(field, false);
        fa.set(object, value);
    }

    /**
     * This method returns every property from the class given as argument. One can tell whether or not he wants the
     * transient properties to be taken into consideration. <br />
     * <b>The properties returned are the ones from the class itself and each of its parent classes.</b>
     * 
     * 
     * @param clazz the class for which to get all properties
     * @param useTransients true if transient fields should be considered, false otherwise
     * @return lhe List of Properties discovered in the class
     */
    public static List<String> getProperties(Class<?> clazz, boolean useTransients) {
        List<String> propertyList = new ArrayList<String>();

        while (clazz.getSuperclass() != null && clazz != Object.class) {
            appendProperties(propertyList, clazz, useTransients);
            clazz = clazz.getSuperclass();
        }

        return propertyList;
    }

    private static void appendProperties(List<String> propertyList, Class<?> clazz, boolean useTransients) {
        for (Method m : ReflectionUtils.getGetters(clazz, useTransients)) {

            String propName = null;
            if (m.getName().startsWith("is")) {
                propName = m.getName().substring(2);
            } else {
                propName = m.getName().substring(3);
            }
            propertyList.add(propName.substring(0, 1).toLowerCase()
                    + (propName.length() <= 1 ? "" : propName.substring(1)));

        }
    }
}
