// ============================================================================
//
// Copyright (c) 2010 - 2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================



package ch.niceideas.common.utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

public class ResourceUtils {

    private static final Logger logger = Logger.getLogger(ResourceUtils.class);
    
    /** Pseudo URL prefix for loading from the class path: "classpath:" */
    public static final String CLASSPATH_URL_PREFIX = "classpath:";
    
    /** URL protocol for an entry from a jar file: "jar" */
    public static final String URL_PROTOCOL_JAR = "jar";
    
    /** URL protocol for a file in the file system: "file" */
    public static final String URL_PROTOCOL_FILE = "file";
    
    /**
     * Search for the filenames in the various classLoader as a Resource and returns an inputStream in it should it
     * be found.
     * 
     * @param fileName the name of the file to look for.
     * @return
     */   
    public static InputStream getResourceAsStream(String fileName) {

        if (StringUtils.isBlank(fileName)) {
            return null;
        }
        
        InputStream inputStream = null;
        try {
            //logger.info("Trying to load configuration from file " + fileName);
            File targetFile = getFile(fileName);
            if (targetFile != null && targetFile.exists()) {
                //logger.info(fileName + " has been found :-) Using it.");
                inputStream = new FileInputStream(targetFile);
            } else {
                logger.debug("Configuration file " + fileName + " does not exist.");
            }
        } catch (FileNotFoundException e) {            
            logger.debug(e, e);
        }
        
        if (fileName.startsWith("classpath:")) {
            fileName = fileName.substring("classpath:".length());
        }

        //logger.info("Now trying to lookup resource " + fileName + " in the classpath.");
        if (inputStream == null) {
            inputStream = ResourceUtils.class.getResourceAsStream(fileName);
        }

        if (inputStream == null) {
            inputStream = ResourceUtils.class.getClassLoader().getResourceAsStream(fileName);
        }

        if (inputStream == null) {
            inputStream = Thread.currentThread().getContextClassLoader().getResourceAsStream(fileName);
        }
        
        if (inputStream == null) {
            logger.warn("Configuration file " + fileName + " was not found.");
        }

        return inputStream;
    }
    
    /**
     * Resolve the given resource location to a <code>java.io.File</code>, i.e. to a file in the file system.
     * <p>
     * Does not check whether the fil actually exists; simply returns the File that the given location would
     * correspond to.
     * 
     * @param resourceLocation the resource location to resolve: either a "classpath:" pseudo URL, a "file:" URL, or
     *            a plain file path
     * @return a corresponding File object
     * @throws FileNotFoundException if the resource cannot be resolved to a file in the file system
     */
    public static File getFile(String resourceLocation) throws FileNotFoundException {
        if (resourceLocation == null) {
            throw new IllegalArgumentException("Resource location must not be null");
        }
        if (resourceLocation.startsWith(CLASSPATH_URL_PREFIX)) {
            String path = resourceLocation.substring(CLASSPATH_URL_PREFIX.length());
            String description = "class path resource [" + path + "]";
            URL url = ClassLoader.getSystemClassLoader().getResource(path);
            if (url == null) {

                url = ResourceUtils.class.getClassLoader().getResource(path);
                if (url == null) {
                    url = Thread.currentThread().getContextClassLoader().getResource(path);

                    if (url == null) {
                        throw new FileNotFoundException(description + " cannot be resolved to absolute file path "
                                + "because it does not reside in the file system");
                    }
                }
            }
            return getFile(url, description);
        }
        try {
            // try URL
            return getFile(new URL(resourceLocation));
        } catch (MalformedURLException ex) {
            // no URL -> treat as file path
            return new File(resourceLocation);
        }
    }
    
    /**
     * Resolve the given resource URL to a <code>java.io.File</code>, i.e. to a file in the file system.
     * 
     * @param resourceUrl the resource URL to resolve
     * @return a corresponding File object
     * @throws FileNotFoundException if the URL cannot be resolved to a file in the file system
     */
    public static File getFile(URL resourceUrl) throws FileNotFoundException {
        return getFile(resourceUrl, "URL");
    }
    
    /**
     * Resolve the given resource URL to a <code>java.io.File</code>, i.e. to a file in the file system.
     * 
     * @param resourceUrl the resource URL to resolve
     * @param description a description of the original resource that the URL was created for (for example, a class
     *            path location)
     * @return a corresponding File object
     * @throws FileNotFoundException if the URL cannot be resolved to a file in the file system
     */
    public static File getFile(URL resourceUrl, String description) throws FileNotFoundException {
        if (resourceUrl == null) {
            throw new IllegalArgumentException("Resource location must not be null");
        }
        
        if (URL_PROTOCOL_JAR.equals(resourceUrl.getProtocol())) {
            return new File(resourceUrl.getFile());
        }
        
        if (!URL_PROTOCOL_FILE.equals(resourceUrl.getProtocol())) {        
            throw new FileNotFoundException(description + " cannot be resolved to absolute file path "
                    + "because it does not reside in the file system: " + resourceUrl);
        }
        try {
            return new File(toURI(resourceUrl).getSchemeSpecificPart());
        } catch (URISyntaxException ex) {
            // Fallback for URLs that are not valid URIs (should hardly ever happen).
            return new File(resourceUrl.getFile());
        }
    }
    

    /**
     * Create a URI instance for the given URL, replacing spaces with "%20" quotes first.
     * <p>
     * Furthermore, this method works on JDK 1.4 as well, in contrast to the <code>URL.toURI()</code> method.
     * 
     * @param url the URL to convert into a URI instance
     * @return the URI instance
     * @throws URISyntaxException if the URL wasn't a valid URI
     * @see java.net.URL#toURI()
     */
    public static URI toURI(URL url) throws URISyntaxException {
        return toURI(url.toString());
    }

    /**
     * Create a URI instance for the given location String, replacing spaces with "%20" quotes first.
     * 
     * @param location the location String to convert into a URI instance
     * @return the URI instance
     * @throws URISyntaxException if the location wasn't a valid URI
     */
    public static URI toURI(String location) throws URISyntaxException {
        return new URI(StringUtils.replace(location, " ", "%20"));
    }
}
