// ============================================================================
//
// Copyright (c) 2010 - 2016, niceideas.ch - Jerome Kehrli
//
// You may distribute this code under the terms of the GNU LGPL license
// (http://www.gnu.org/licenses/lgpl.html). [^]
//
// ===================================================================

package ch.niceideas.common.utils;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;

import org.apache.log4j.Logger;

/**
 * Various utility methods aimed at helping with stream handling.
 */
public abstract class StreamUtils {

    private static final Logger logger = Logger.getLogger(StreamUtils.class);

    /**
     * Default value is 2048.
     */
    public static final int DEFAULT_BUFFER_SIZE = 2048;

    /**
     * Copies information from the input stream to the output stream using the specified buffer size
     * 
     * @throws java.io.IOException
     */
    public static void copy(InputStream input, OutputStream output, int bufferSize) throws IOException {
        if (input != null) {
            byte[] buf = new byte[bufferSize];
            int bytesRead = input.read(buf);
            while (bytesRead != -1) {
                output.write(buf, 0, bytesRead);
                bytesRead = input.read(buf);
            }
            output.flush();
        }
    }

    /**
     * Copy chars from a <code>Reader</code> to a <code>Writer</code>.
     * 
     * @param input the <code>Reader</code> to read from
     * @param output the <code>Writer</code> to write to
     * @return the number of characters copied
     * @throws NullPointerException if the input or output is null
     * @throws IOException if an I/O error occurs
     */
    public static long copy(Reader input, Writer output) throws IOException {
        char[] buffer = new char[DEFAULT_BUFFER_SIZE];
        long count = 0;
        int n = 0;
        while (-1 != (n = input.read(buffer))) {
            output.write(buffer, 0, n);
            count += n;
        }
        return count;
    }

    /**
     * Copy chars from a <code>InputStream</code> to a <code>Writer</code>.
     * 
     * @param is
     * @param output
     * @return
     * @throws IOException
     */
    public static long copy(InputStream is, Writer output) throws IOException {
        return copy(new InputStreamReader(is), output);
    }

    /**
     * Copies information from the input stream to the output stream using a default buffer size of 2048 bytes.
     * 
     * @throws java.io.IOException
     */
    public static void copy(InputStream input, OutputStream output) throws IOException {
        copy(input, output, DEFAULT_BUFFER_SIZE);
    }

    /**
     * Copies information between specified streams and then closes both of the streams.
     * 
     * @throws java.io.IOException
     */
    public static void copyThenClose(InputStream input, OutputStream output) throws IOException {
        copy(input, output);
        input.close();
        output.close();
    }

    /**
     * @return a byte[] containing the information contained in the specified InputStream.
     * @throws java.io.IOException
     */
    public static byte[] getBytes(InputStream input) throws IOException {
        if (input == null) {
            return null;
        }
        ByteArrayOutputStream result = new ByteArrayOutputStream();        
        copy(input, result);
        result.close();
        return result.toByteArray();
    }

    /**
     * Get the Input stream as a String using the given encoding
     * 
     * @param input the InputStream
     * @param Strin encoding the stream encoding
     * @return the input stream as a string
     */
    public static String getAsString(InputStream input, String encoding) throws IOException {

        if (input == null) {
            return "";
        }

        StringBuilder sb = new StringBuilder();
        BufferedReader d = new BufferedReader(encoding == null ? new InputStreamReader(input) : new InputStreamReader(
                input, encoding));

        try {
            String line = null;
            while ((line = d.readLine()) != null) {
                sb.append(line + "\n");
            }
        } catch (IOException e) {
            logger.error("getAsString", e);
            throw (e);
            
        } finally {
        	close (input);
        }

        String retValue = sb.toString();

        return retValue;

    }

    /**
     * Get the Input stream as a String
     * 
     * @param input the InputStream
     * @return the input stream as a string
     */
    public static String getAsString(InputStream input) throws IOException {
        return getAsString(input, null);
    }

    /**
     * Close the <code> in </code> stream and ignored any exception.
     * 
     * @param in may be null (ignored)
     */
    public static void close(InputStream in) {
        if (in != null) {
            try {
                in.close();
            } catch (Throwable ignored) {
            }
        }
    }

    /**
     * Close the <code> out </code> stream and ignored any exception.
     * 
     * @param in may be null (ignored)
     */
    public static void close(OutputStream out) {
        if (out != null) {
            try {
                out.close();
            } catch (Throwable ignored) {
            }
        }
    }

    /**
     * Close the <code> in </code> reader and ignored any exception.
     * 
     * @param in may be null (ignored)
     */
    public static void close(Reader in) {
        if (in != null) {
            try {
                in.close();
            } catch (Throwable ignored) {
            }
        }
    }

    /**
     * Close the <code> out </code> writer and ignored any exception.
     * 
     * @param out may be null (ignored)
     */
    public static void close(Writer out) {
        if (out != null) {
            try {
                out.close();
            } catch (Throwable ignored) {
            }
        }
    }

}
